import os
import random
from typing import List, Optional, Tuple

import numpy as np
import torch


def show_environment_setting(print_fn):
    msg = "CUDA INFO:"
    msg += f"\n\tIs CUDA supported by this system? {torch.cuda.is_available()}"
    if torch.cuda.is_available():
        msg += f"\n\tCUDA version: {torch.version.cuda}"
        msg += f"\n\tID of current CUDA device: {torch.cuda.current_device()}"
        msg += f"\n\tName of current CUDA device: {torch.cuda.get_device_name(torch.cuda.current_device())}"
    print_fn(msg)

def att_mask_from_individual_masks(tgt_mask: torch.Tensor, src_mask: torch.Tensor)-> torch.Tensor:
    """Combine two individual masks for attention mask

    :param tgt_mask: mask for target sequence with shape (bsz, tgt_len)
    :type tgt_mask: torch.Tensor
    :param src_mask: mask for source sequence with shape (bsz, src_len)
    :type src_mask: torch.Tensor
    :return: attention mask with shape (bsz, tgt_len, src_len)
    :rtype: torch.Tensor
    """
    tgt_len = tgt_mask.shape[-1]
    src_len = src_mask.shape[-1]
    
    _tgt_mask = tgt_mask.unsqueeze(-1).repeat(1, 1, src_len)
    _src_mask = src_mask.unsqueeze(1).repeat(1, tgt_len, 1)
    
    return _tgt_mask * _src_mask

def zero_pad_batching_one_dim(tensor_list: List[torch.Tensor], return_mask=True) -> Tuple[torch.Tensor, torch.Tensor]:
    max_len = max([len(tensor) for tensor in tensor_list])
    device = tensor_list[0].device
    shape = (len(tensor_list), max_len) if len(tensor_list[0].shape) == 1 else (len(tensor_list), max_len, *list(tensor_list[0].shape[1:]))
    token_tensor = torch.zeros(shape, dtype=tensor_list[0].dtype, device=device)
    if return_mask:
        mask_tensor = torch.ones(token_tensor.shape[:-1], dtype=torch.bool, device=device)
    for idx, tensor in enumerate(tensor_list):
        token_tensor[idx, :len(tensor)] = tensor
        if return_mask:
            mask_tensor[idx, :len(tensor)] = False
    return (token_tensor, mask_tensor) if return_mask else token_tensor

def zero_pad_batching_two_dim(tensor_list: List[torch.Tensor], return_mask=True) -> torch.Tensor:
    device = tensor_list[0].device
    dim_one_max_len = max([len(tensor) for tensor in tensor_list])
    dim_two_max_len = max([len(tensor[0]) for tensor in tensor_list])
    token_tensor = torch.zeros(len(tensor_list), dim_one_max_len, dim_two_max_len, dtype=tensor_list[0].dtype, device=device)
    if return_mask:
        mask_tensor = torch.ones(token_tensor, dtype=torch.bool, device=device)
    for idx, tensor in enumerate(tensor_list):
        token_tensor[idx, :len(tensor), :len(tensor[0])] = tensor
        if return_mask:
            mask_tensor[idx, :len(tensor), :len(tensor[0])] = False
    return (token_tensor, mask_tensor) if return_mask else token_tensor

def zero_pad_batching(tensor_list: List[torch.Tensor]) -> torch.Tensor:
    tensor_item = tensor_list[0]
    if len(tensor_item.shape) == 1:
        return zero_pad_batching_one_dim(tensor_list)
    elif len(tensor_item.shape) == 2:
        return zero_pad_batching_two_dim(tensor_list)
    else:
        raise NotImplementedError(f"Only support 1D and 2D tensor, but found: {tensor_item.shape}")

def set_torch_deterministic(seed: Optional[int]=0):
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)  # if you are using multi-GPU.
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    
if __name__ == "__main__":
    show_environment_setting()




import unittest

import src.hkkang_utils.io as io_utils

class Test_io(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_io, self).__init__(*args, **kwargs)
    
    def _inner_func(self, *args, **kwargs):
        print(f"Inner function called with args: {args} and kwargs: {kwargs}")
    
    def test_io(self):
        output = io_utils.intercept_stdout(self._inner_func)(1, 2, 3, a=4, b=5, c=6)
        self.assertTrue(output.startswith("Inner function called with args: "))
        print(f"Test passed! Output: {output}")

if __name__ == "__main__":
    unittest.main()



from typing import Any, List, Tuple, Union


def compute_f1(preds: List[Any], labels: List[Any], compare_func: Any=lambda *k: k[0] == k[1], return_cnt=False) -> Union[Tuple[float, float, float], Tuple[float, float, float, float, float, float]]:
    """Compute F1 score for between two lists of items. Order of the items in the lists does not matter."""
    tp, fp, fn = 0, 0, 0
    # Count tp, fp, fn
    label_indices = [i for i in range(len(labels))]
    for pred in preds:
        found = False
        # Check if pred is in labels
        for i in label_indices:
            if compare_func(pred, labels[i]):
                tp += 1
                label_indices.remove(i)
                found= True
                break
        if not found:
            fp += 1
    fn = len(label_indices)
    # Compute f1, precision, recall
    precision = tp / (tp + fp)
    recall = tp / (tp + fn)
    f1 = 2 * precision * recall / (precision + recall)
    if return_cnt:
        return f1, precision, recall, tp, fp, fn
    else:
        return f1, precision, recall


import unittest

import src.hkkang_utils.list as list_utils


class Test_list_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_list_utils, self).__init__(*args, **kwargs)
    
    def test_do_flatten_list(self):
        input_list = [[1,2,3], [4,5,6], [7,8,9]]
        gold = [1,2,3,4,5,6,7,8,9]
        result = list_utils.do_flatten_list(input_list)
        self.assertEqual(result, gold, f"result: {result}, gold_after: {gold}")
        
    
    def test_map_many(self):
        input_list = [1,2,3,4,5]
        functions = [lambda x: x+1, lambda x: x*2]
        gold = [4,6,8,10,12]
        result = list_utils.map_many(functions, input_list)
        self.assertEqual(result, gold, f"result: {result}, gold: {gold}")


if __name__ == "__main__":
    unittest.main()
    
    


__version__ ="0.0.1"
__all__ = ["concurrent", "file", "io", "list", "logging", "metrics", "misc", "ml", "pg", "sql", "string", "tensor", "testing", "time"]

import hkkang_utils.concurrent
import hkkang_utils.file
import hkkang_utils.io
import hkkang_utils.list
import hkkang_utils.logging
import hkkang_utils.metrics
import hkkang_utils.misc
import hkkang_utils.ml
import hkkang_utils.pg
import hkkang_utils.sql
import hkkang_utils.string
import hkkang_utils.tensor
import hkkang_utils.testing
import hkkang_utils.time



import os
import unittest

import src.hkkang_utils.testing as test_utils

TEST_DIR_PATH = "test_dir"


class Test_test_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_test_utils, self).__init__(*args, **kwargs)
    
    @test_utils.set_and_clean_test_dir(TEST_DIR_PATH)
    def _func_to_be_decorated(self):
        # Check if directory exists
        self.assertTrue(os.path.exists(TEST_DIR_PATH))
    
    def test_decorator(self):
        # Make sure that the directory does not exists
        self.assertFalse(os.path.exists(TEST_DIR_PATH))
        # Check if the decorator works
        self._func_to_be_decorated()
        # Check if directory is removed
        self.assertFalse(os.path.exists(TEST_DIR_PATH))

if __name__ == "__main__":
    unittest.main()



import unittest

import src.hkkang_utils.metrics as metrics_utils


class Test_metrics_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_metrics_utils, self).__init__(*args, **kwargs)
    
    def test_compute_f1(self):
        # test_data
        test_data = [{"pred": [1,2,3,4,5,6], "gold": [1,2,3,4,5,6,7,8,9,10], "f1": 0.7499999999999999, "precision": 1, "recall": 0.6},
                     {"pred": [1,2,3,4,5,6,7,8,9,10], "gold": [1,2,3,4,5,6,7,8,9,10], "f1": 1, "precision": 1, "recall": 1},
                     {"pred": [1,2,3,4,5,6,7,8,9,10,11,12], "gold": [1,2,3,4,5,6,7,8,9,10], "f1": 0.9090909090909091, "precision": 10/12, "recall": 1}]
        for datum in test_data:
            f1, precision, recall = metrics_utils.compute_f1(datum["pred"],datum["gold"])
            self.assertEqual(recall, datum["recall"], f"recall pred: {recall}, gold: {datum['recall']}")
            self.assertEqual(precision, datum["precision"], f"precision pred: {precision}, gold: {datum['precision']}")
            self.assertEqual(f1, datum["f1"], f"f1 pred: {f1}, gold: {datum['f1']}")
    
if __name__ == "__main__":
    unittest.main()
    
    